/*
 * PDFlib PLOP DS:
 * generate a dynamic visualization document in memory with PDFlib,
 * and digitally sign a second document, using the dynamic document
 * for signature visualization.
 */

#include <iostream>
#include <iomanip>
#include <string>
#include <time.h>

#include "pdflib.hpp"
#include "plop.hpp"

using namespace std;
using namespace pdflib;

namespace
{
wstring get_wstring(const PLOP& plop, const string& utf8_string);
}

int
main(int argc, char *argv[])
{
    const char *contents = NULL;
    long size;
    PDFlib p;

    const wchar_t * const months[] =
    {
        L"January", L"February", L"March", L"April", L"May", L"June",
        L"July", L"August", L"September", L"October", L"November", L"December"
    };

    /* This is where input files live. Adjust as necessary. */
    const wstring searchpath(L"../data");

    /* Basic signature options */
    static const wstring sign_opts =
        L"ltv=try " /* use ltv=full to ensure LTV status */
#if defined(USE_PKCS11)
        L"engine=pkcs#11 digitalid={filename=cryptoki.dll}"
#else
        L"engine=builtin digitalid={filename=demo_signer_rsa_2048.p12} password={demo}"
#endif
    ;

    if (argc < 3)
    {
        wcerr << L"usage: dynamicsign <infile> <outfile>" << endl;
        return 2;
    }

    /* ---------- Step 1: create visualization document in memory ---------- */

    try
    {
        /* create a new PDF document in memory */
        p.begin_document(L"", L"");

        p.set_info(L"Creator", L"hellosign.cpp");
        p.set_info(L"Author", L"Rainer Pl\u00f6ckl");
        p.set_info(L"Title", L"Dynamic signature visualization document");

        p.begin_page_ext(300, 50, L"");

        time_t timer;
        struct tm ltime;

        time(&timer);
        ltime = *localtime(&timer);

        wostringstream timebuf;
        timebuf << L"Digitally signed "
                << months[ltime.tm_mon] << L" " << ltime.tm_mday
                << L", " << ltime.tm_year + 1900
                << L" at "
                << setw(2) << setfill(L'0') << ltime.tm_hour << L":"
                << setw(2) << setfill(L'0') << ltime.tm_min << L":"
                << setw(2) << setfill(L'0') << ltime.tm_sec;

        wstring const textoptlist(L"fontsize=10 fontname=Helvetica "
            L"encoding=unicode fillcolor=red");

        p.fit_textline(timebuf.str(), 10, 10, textoptlist);
        p.end_page_ext(L"");

        p.end_document(L"");

        /* fetch the memory buffer containing the PDF document */
        contents = p.get_buffer(&size);
    }
    catch (PDFlib::Exception &ex)
    {
        wcerr << L"PDFlib exception occurred in hellosign sample:" << endl;
        wcerr << L"[" << ex.get_errnum() << L"] " << ex.get_apiname() << L": "
                << ex.get_errmsg() << endl;

        return 98;
    }

    /* ------------------ Step 2: sign the document ----------------- */

    try
    {
        int doc;
        PLOP plop;
        wostringstream optlist;

        optlist << L"searchpath={" << searchpath << L"}";

        plop.set_option(optlist.str());

        /* Create a virtual PVF file from the document data in memory */
        plop.create_pvf(L"/pvf/visdoc.pdf", contents, size, L"");

        /*
         * Caution: For simplicity we assume that the program arguments are
         * encoded as UTF-8, which might not be true in all cases!
         */
        const wstring indoc_name(get_wstring(plop, argv[1]));

        /* open input document from PVF file */
        if ((doc = plop.open_document(indoc_name, L"")) == -1)
        {
            wcerr << L"Error: " << plop.get_apiname() << L": "
                    << plop.get_errmsg() << endl;
            return 99;
        }

        int visdoc = plop.open_document(L"/pvf/visdoc.pdf", L"");
        if (visdoc == -1)
        {
            wcerr << L"Error: " << plop.get_apiname() << L": "
                    << plop.get_errmsg() << endl;
            return 99;
        }

        /* Add field option for visualization to base signature options */
        optlist.str(L"");
        optlist << sign_opts << L" field={visdoc=" << visdoc
                << L" rect={10 10 adapt adapt}} ";
        if (plop.prepare_signature(optlist.str()) == -1)
        {
            wcerr << L"Error: " << plop.get_apiname() << L": "
                    << plop.get_errmsg() << endl;
            return 2;
        }

        /* sign the output file */
        optlist.str(L"");
        optlist << L"input=" << doc;
        if (plop.create_document(get_wstring(plop, argv[2]), optlist.str())
                                                    == -1)
        {
            wcerr << L"Error: " << plop.get_apiname() << L": "
                    << plop.get_errmsg() << endl;
            return 99;
        }

        /* close the input document */
        plop.close_document(doc, L"");

        /* close the visualization document */
        plop.close_document(visdoc, L"");

        /* delete PVF file containing input document */
        if (plop.delete_pvf(L"/pvf/visdoc.pdf") == -1)
        {
            wcerr << L"Error: " << plop.get_apiname() << L": "
                    << plop.get_errmsg() << endl;
            return 99;
        }
    }
    catch (PLOP::Exception &ex)
    {
        wcerr << L"PLOP exception occurred in hellosign:" << endl;
        wcerr << L"[" << ex.get_errnum() << L"] " << ex.get_apiname() << L": "
                << ex.get_errmsg() << endl;
        return 99;
    }

    return 0;
}

namespace
{

/*
 * Get a wstring for the given string, assuming the string is encoded in
 * UTF-8.
 */
wstring get_wstring(const PLOP& plop, const string& utf8_string)
{
    const size_t size = sizeof(wstring::value_type);
    string wide_string;

    switch (size)
    {
    case 2:
        wide_string = plop.convert_to_unicode(L"utf8",
                            utf8_string, L"outputformat=utf16");
        break;

    case 4:
        wide_string = plop.convert_to_unicode(L"utf8",
                            utf8_string, L"outputformat=utf32");
        break;

    default:
        throw std::logic_error("Unsupported wchar_t size");
    }

    return wstring(reinterpret_cast<const wchar_t *>(wide_string.data()),
            wide_string.length() / size);
}

} // end anonymous namespace
